@@ -40,6 +40,8 @@ from flask import (
     stream_template, stream_template_string, stream_with_context,
 )
 
+from werkzeug.datastructures import FileStorage
+
 import logging
 log = logging.getLogger("fsa")
 
@@ -56,7 +58,7 @@ ObjectPermsFun = Callable[[str, Any, Optional[str]], bool]
 PasswordCheckFun = Callable[[str, str], Optional[bool]]
 PasswordQualityFun = Callable[[str], bool]
 CastFun = Callable[[str], object]
-SpecialParameterFun = Callable[[], Any]
+SpecialParameterFun = Callable[[str], Any]
 HeaderFun = Callable[[Response], Optional[str]]
 BeforeRequestFun = Callable[[Request], Optional[Response]]
 AfterRequestFun = Callable[[Response], Response]
@@ -178,6 +180,12 @@ def _typeof(p: inspect.Parameter):
         return str
 
 
+def _get_file_storage(p: str) -> FileStorage:
+    if p not in request.files:
+        raise ErrorResponse(f"missing file parameter \"{p}\"", 400)
+    return request.files[p]
+
+
 class Reference(ppp.Proxy):
     """Convenient object wrapper class."""
 
@@ -249,7 +257,7 @@ _DIRECTIVES = {
     "FSA_MODE", "FSA_LOGGING_LEVEL",
     # general settings
     "FSA_SECURE", "FSA_SERVER_ERROR", "FSA_NOT_FOUND_ERROR", "FSA_LOCAL",
-    "FSA_HANDLE_ALL_ERRORS",
+    "FSA_HANDLE_ALL_ERRORS", "FSA_KEEP_USER_ERRORS",
     # register hooks
     "FSA_GET_USER_PASS", "FSA_USER_IN_GROUP", "FSA_CAST", "FSA_OBJECT_PERMS",
     "FSA_SPECIAL_PARAMETER", "FSA_ERROR_RESPONSE", "FSA_ADD_HEADERS",
@@ -280,6 +288,7 @@ _DEFAULT_CACHE_TTL = 600  # seconds, 10 minutes
 _DEFAULT_REJECT_UNEXPECTED_PARAM = True
 _DEFAULT_SERVER_ERROR = 500
 _DEFAULT_NOT_FOUND_ERROR = 404
+_DEFAULT_KEEP_USER_ERRORS = False
 _DEFAULT_ERROR_RESPONSE = "plain"
 _DEFAULT_PASSWORD_SCHEME = "bcrypt"
 _DEFAULT_PASSWORD_OPTS = {"bcrypt__default_rounds": 4, "bcrypt__default_ident": "2y"}
@@ -313,12 +322,13 @@ class FlaskSimpleAuth:
             JsonData: json.loads,
         }
         self._special_parameters: Dict[type, SpecialParameterFun] = {
-            Request: lambda: request,
-            Environ: lambda: request.environ,
-            Session: lambda: session,
-            Globals: lambda: g,
-            CurrentUser: lambda: self.current_user(),
-            CurrentApp: lambda: current_app,
+            Request: lambda _: request,
+            Environ: lambda _: request.environ,
+            Session: lambda _: session,
+            Globals: lambda _: g,
+            CurrentUser: lambda _: self.current_user(),
+            CurrentApp: lambda _: current_app,
+            FileStorage: _get_file_storage,
         }
         self._auth: List[str] = []
         self._http_auth = None
@@ -328,9 +338,10 @@ class FlaskSimpleAuth:
         # fsa-generated errors
         self._server_error: int = _DEFAULT_SERVER_ERROR
         self._not_found_error: int = _DEFAULT_NOT_FOUND_ERROR
+        self._keep_user_errors: bool = _DEFAULT_KEEP_USER_ERRORS
         self._secure: bool = True
         self._secure_warning = True
-        self._names: Set[str] = set()
+        self._auth_params: Set[str] = set()
         self._groups: Set[Union[str, int]] = set()
         self._scopes: Set[str] = set()
         self._headers: Dict[str, Union[HeaderFun, str]] = {}
@@ -353,13 +364,18 @@ class FlaskSimpleAuth:
         assert self._error_response is not None
         return self._error_response(msg, code)
 
+    def _Exc(self, exc):
+        """Handle a user error."""
+        if exc and not hasattr(exc, "_fsa_traced"):
+            log.error(exc, exc_info=True)
+            setattr(exc, "_fsa_traced", True)
+        return exc if self._keep_user_errors else None
+
     def _Err(self, msg: str, code: int, exc: Exception = None):
         """Build and trace an ErrorResponse exception with a message."""
         if self._mode >= Mode.DEBUG3:
             log.debug(f"error: {code} {msg}")
-        if exc:
-            log.info(exc, exc_info=True)
-        return ErrorResponse(msg, code)
+        return self._Exc(exc) or ErrorResponse(msg, code)
 
     def _Bad(self, msg: str):
         """Build and trace an exception on a bad configuration."""
@@ -391,13 +407,17 @@ class FlaskSimpleAuth:
 
     def _auth_reset_user(self):
         """Before request hook to cleanup authentication and authorization."""
-        self._local.routed = False
-        self._local.source = None
-        self._local.user = None
-        self._local.need_authorization = True
-        self._local.auth = self._auth
-        self._local.scopes = None
         self._local.start = dt.datetime.timestamp(dt.datetime.now())
+        # whether some routing has occurred, vs a before_request generated response
+        self._local.routed = False
+        # authentication and authorizations
+        self._local.source = None              # what authn has been used
+        self._local.user = None                # for this user
+        self._local.need_authorization = True  # whether some authz occurred
+        self._local.auth = self._auth          # allowed authn schemes
+        self._local.scopes = None              # current oauth scopes
+        # parameters
+        self._local.params = None              # json|http
 
     def _run_before_requests(self):
         """Run internal before request hooks."""
@@ -496,8 +516,11 @@ class FlaskSimpleAuth:
     def _params(self):
         """Get request parameters wherever they are."""
         if request.is_json:
+            self._local.params = "json"
             return request.json
         else:
+            self._local.params = "http"
+
             # reimplement "request.values" after Flask 2.0 regression
             # the logic of web-targetted HTTP does not make sense for a REST API
             # https://github.com/pallets/werkzeug/pull/2037
@@ -641,6 +664,7 @@ class FlaskSimpleAuth:
         elif local == "thread":
             from threading import local as Local  # type: ignore
         elif local == "werkzeug":
+            # FIXME should handle various *let stuff
             from werkzeug.local import Local  # type: ignore
         else:
             raise self._Bad(f"unexpected FSA_LOCAL value: {local}")
@@ -652,7 +676,13 @@ class FlaskSimpleAuth:
         # status code for some errors errors
         self._server_error = conf.get("FSA_SERVER_ERROR", _DEFAULT_SERVER_ERROR)
         self._not_found_error = conf.get("FSA_NOT_FOUND_ERROR", _DEFAULT_NOT_FOUND_ERROR)
-        # actual error response generation
+        # error response generation
+        if conf.get("FSA_HANDLE_ALL_ERRORS", True):
+            # take responsability for handling errors
+            self._app.register_error_handler(exceptions.HTTPException, lambda e: self._Res(e.description, e.code))
+        # override FSA internal error handling user errors
+        self._keep_user_errors = conf.get("FSA_KEEP_USER_ERRORS", False)
+        # how to generate an error response
         if self._error_response is None:
             error = conf.get("FSA_ERROR_RESPONSE", _DEFAULT_ERROR_RESPONSE)
             if error is None:
@@ -672,9 +702,6 @@ class FlaskSimpleAuth:
                 raise self._Bad(f"unexpected FSA_ERROR_RESPONSE value: {error}")
         elif "FSA_ERROR_RESPONSE" in conf:
             log.warning("ignoring FSA_ERROR_RESPONSE directive, handler already set")
-        # possibly take responsability for handling errors
-        if conf.get("FSA_HANDLE_ALL_ERRORS", True):
-            self._app.register_error_handler(exceptions.HTTPException, lambda e: self._Res(e.description, e.code))
         # whether to error on unexpected parameters
         self._reject_param = \
             conf.get("FSA_REJECT_UNEXPECTED_PARAM", _DEFAULT_REJECT_UNEXPECTED_PARAM)
@@ -706,7 +733,11 @@ class FlaskSimpleAuth:
         self._cors: bool = conf.get("FSA_CORS", False)
         self._cors_opts: Dict[str, Any] = conf.get("FSA_CORS_OPTS", {})
         if self._cors:
-            from flask_cors import CORS  # type: ignore
+            try:
+                from flask_cors import CORS  # type: ignore
+            except ModuleNotFoundError:  # pragma: no cover
+                log.error("missing module: install FlaskSimpleAuth[cors]")
+                raise
             CORS(self._app, **self._cors_opts)
         self._401_redirect: Optional[str] = conf.get("FSA_401_REDIRECT", None)
         self._url_name: Optional[str] = conf.get("FSA_URL_NAME", "URL" if self._401_redirect else None)
@@ -716,6 +747,7 @@ class FlaskSimpleAuth:
         self._cache_opts: Dict[str, Any] = conf.get("FSA_CACHE_OPTS", {})
         cache = conf.get("FSA_CACHE", _DEFAULT_CACHE)
         if cache:
+            # NOTE no try/except because the dependency is mandatory
             import cachetools as ct
             import CacheToolsUtils as ctu  # type: ignore
 
@@ -745,7 +777,11 @@ class FlaskSimpleAuth:
                 self._cache = ctu.StatsCache(rcache)
                 self._gen_cache = ctu.PrefixedCache
             elif cache in ("memcached", "pymemcache"):
-                import pymemcache as pmc  # type: ignore
+                try:
+                    import pymemcache as pmc  # type: ignore
+                except ModuleNotFoundError:  # pragma: no cover
+                    log.error("missing module: install FlaskSimpleAuth[memcached]")
+                    raise
 
                 if "serde" not in self._cache_opts:
                     self._cache_opts.update(serde=ctu.JsonSerde())
@@ -754,7 +790,11 @@ class FlaskSimpleAuth:
                 self._cache = ctu.StatsMemCached(pmc.Client(**self._cache_opts))
                 self._gen_cache = ctu.PrefixedMemCached
             elif cache == "redis":
-                import redis
+                try:
+                    import redis
+                except ModuleNotFoundError:  # pragma: no cover
+                    log.error("missing module: install FlaskSimpleAuth[redis]")
+                    raise
 
                 ttl = self._cache_opts.pop("ttl", _DEFAULT_CACHE_TTL)
                 rc = redis.Redis(**self._cache_opts)
@@ -791,7 +831,6 @@ class FlaskSimpleAuth:
             raise self._Bad(f"Token carrier {self._carrier} requires a name")
         if self._carrier == "param":
             assert isinstance(self._name, str)
-            self._names.add(self._name)
         # token realm and possible issuer…
         realm = conf.get("FSA_REALM", self._app.name)
         if self._token == "fsa":  # simplify realm for fsa
@@ -861,7 +900,15 @@ class FlaskSimpleAuth:
         self._login = conf.get("FSA_FAKE_LOGIN", "LOGIN")
         self._userp = conf.get("FSA_PARAM_USER", "USER")
         self._passp = conf.get("FSA_PARAM_PASS", "PASS")
-        self._names.update([self._login, self._userp, self._passp])
+        # parameters used for authn
+        if "param" in self._auth or "password" in self._auth:
+            self._auth_params.add(self._userp)
+            self._auth_params.add(self._passp)
+        if "fake" in self._auth:
+            self._auth_params.add(self._login)
+        if "token" in self._auth and self._carrier == "param":
+            assert isinstance(self._name, str)
+            self._auth_params.add(self._name)
         #
         # authentication and authorization hooks
         #
@@ -886,7 +933,11 @@ class FlaskSimpleAuth:
         #
         if self._auth_has("http-basic", "http-digest", "http-token", "digest"):
             opts = conf.get("FSA_HTTP_AUTH_OPTS", {})
-            import flask_httpauth as fha  # type: ignore
+            try:
+                import flask_httpauth as fha  # type: ignore
+            except ModuleNotFoundError:  # pragma: no cover
+                log.error("missing module: install FlaskSimpleAuth[httpauth]")
+                raise
 
             if "http-basic" in self._auth:
                 self._http_auth = fha.HTTPBasicAuth(realm=self._realm, **opts)
@@ -972,7 +1023,11 @@ class FlaskSimpleAuth:
             # passlib context is a pain, you have to know the scheme name to set its
             # round. Ident "2y" is same as "2b" but apache compatible.
             options = conf.get("FSA_PASSWORD_OPTS", _DEFAULT_PASSWORD_OPTS)
-            from passlib.context import CryptContext  # type: ignore
+            try:
+                from passlib.context import CryptContext  # type: ignore
+            except ModuleNotFoundError:  # pragma: no cover
+                log.error("missing module: install FlaskSimpleAuth[passwords]")
+                raise
 
             self._pm = CryptContext(schemes=[scheme], **options)
 
@@ -997,7 +1052,7 @@ class FlaskSimpleAuth:
         params = self._params()
         user = params.get(self._login, None)
         if not user:
-            raise self._Err(f"missing login parameter: {self._login}", 401)
+            raise self._Err(f"missing fake login parameter: {self._login}", 401)
         return user
 
     #
@@ -1056,7 +1111,8 @@ class FlaskSimpleAuth:
                 raise e
             except Exception as e:
                 log.info(f"AUTH (hook) failed: {e}")
-                log.info(e, exc_info=True)
+                if self._Exc(e):  # pragma: no cover
+                    raise
                 return False
         return False
 
@@ -1118,16 +1174,15 @@ class FlaskSimpleAuth:
     #
     def _get_basic_auth(self):
         """Get user with basic authentication."""
-        import base64 as b64
-
         auth = request.headers.get("Authorization", None)
         if not auth:
-            log.debug("AUTH (basic): missing authorization header")
+            log.debug("AUTH (basic): missing Authorization header")
             raise self._Err("missing authorization header", 401)
         if not auth.startswith("Basic "):
             log.debug(f'AUTH (basic): unexpected auth "{auth}"')
-            raise self._Err("unexpected authorization header", 401)
+            raise self._Err("unexpected Authorization header", 401)
         try:
+            import base64 as b64
             user, pwd = b64.b64decode(auth[6:]).decode().split(":", 1)
         except Exception as e:
             log.debug(f'AUTH (basic): error while decoding auth "{auth}" ({e})')
@@ -1147,10 +1202,10 @@ class FlaskSimpleAuth:
         params = self._params()
         user = params.get(self._userp, None)
         if not user:
-            raise self._Err(f"missing login parameter: {self._userp}", 401)
+            raise self._Err(f"missing param login parameter: {self._userp}", 401)
         pwd = params.get(self._passp, None)
         if not pwd:
-            raise self._Err(f"missing password parameter: {self._passp}", 401)
+            raise self._Err(f"missing param password parameter: {self._passp}", 401)
         return self._check_password(user, pwd)
 
     #
@@ -1224,15 +1279,20 @@ class FlaskSimpleAuth:
         - iss: issuer (the source)
         - aud : audience (the realm)
         - not used: iat (issued at), nbf (not before), jti (jtw id)
-        - scope: optional authorizations (for testing)
+        - scope: optional authorizations
         """
-        exp = dt.datetime.now(dt.timezone.utc) + dt.timedelta(minutes=delay)
-        import jwt
+        try:
+            import jwt
+        except ModuleNotFoundError:  # pragma: no cover
+            log.error("missing module: install FlaskSimpleAuth[jwt]")
+            raise
 
+        exp = dt.datetime.now(dt.timezone.utc) + dt.timedelta(minutes=delay)
         token = {"exp": exp, "sub": user, "aud": realm}
         if issuer:
             token.update(iss=issuer)
         if scope:
+            # NOTE Why doesn't JWT use a list there?
             token.update(scope=" ".join(scope))
         return jwt.encode(token, secret, algorithm=self._algo)
 
@@ -1243,7 +1303,7 @@ class FlaskSimpleAuth:
         )
 
     def create_token(self, user: str = None, realm: str = None,
-                     issuer: str = None, delay: float = None):
+                     issuer: str = None, delay: float = None, **kwargs):
         """Create a new token for user depending on the configuration."""
         assert self._token
         user = user or self.get_user()
@@ -1251,43 +1311,52 @@ class FlaskSimpleAuth:
         issuer = issuer or self._issuer
         delay = delay or self._delay
         return (
-            self._get_fsa_token(realm, issuer, user, delay, self._secret) if self._token == "fsa" else
-            self._get_jwt_token(realm, issuer, user, delay, self._sign)
+            self._get_fsa_token(realm, issuer, user, delay, self._secret, **kwargs) if self._token == "fsa" else
+            self._get_jwt_token(realm, issuer, user, delay, self._sign, **kwargs)
         )
 
     def _get_fsa_token_auth(self, token):
         """Tell whether FSA token is ok: return validated user or None."""
         # token format: "realm[/issuer]:calvin:20380119031407:<signature>"
         if token.count(":") != 3:
-            log.debug(f"AUTH (fsa token): unexpected token {token}")
+            if self._mode >= Mode.DEBUG:
+                log.debug(f"AUTH (fsa token): unexpected token ({token})")
             raise self._Err(f"invalid fsa token: {token}", 401)
         realm, user, slimit, sig = token.split(":", 3)
         try:
             limit = self._from_timestamp(slimit)
         except Exception as e:
-            log.debug(f"AUTH (fsa token): malformed timestamp {slimit}: {e}")
-            raise self._Err(f"unexpected limit: {slimit}", 401, e)
+            if self._mode >= Mode.DEBUG:
+                log.debug(f"AUTH (fsa token): malformed timestamp {slimit} ({token}): {e}")
+            raise self._Err(f"unexpected fsa token limit: {slimit} ({token})", 401, e)
         # check realm
         if self._issuer and realm != f"{self._realm}/{self._issuer}" or \
            not self._issuer and realm != self._realm:
-            log.debug(f"AUTH (fsa token): unexpected realm {realm}")
-            raise self._Err(f"unexpected realm: {realm}", 401)
+            if self._mode >= Mode.DEBUG:
+                log.debug(f"AUTH (fsa token): unexpected realm {realm} ({token})")
+            raise self._Err(f"unexpected fsa token realm: {realm} ({token})", 401)
         # check signature
         ref = self._cmp_sig(f"{realm}:{user}:{slimit}", self._secret)
         if ref != sig:
-            log.debug("AUTH (fsa token): invalid signature")
-            raise self._Err("invalid fsa auth token signature", 401)
+            if self._mode >= Mode.DEBUG:
+                log.debug("AUTH (fsa token): invalid signature ({token})")
+            raise self._Err("invalid fsa auth token signature ({token})", 401)
         # check limit with a grace time
         now = dt.datetime.now(dt.timezone.utc) - dt.timedelta(minutes=self._grace)
         if now > limit:
-            log.debug("AUTH (fsa token): token {token} has expired")
-            raise self._Err("expired fsa auth token", 401)
+            if self._mode >= Mode.DEBUG:
+                log.debug("AUTH (fsa token): token {token} has expired ({token})")
+            raise self._Err("expired fsa auth token ({token})", 401)
         # all is well
         return user, limit, None
 
     def _get_jwt_token_auth(self, token):
         """Tell whether JWT token is ok: return validated user or None."""
-        import jwt
+        try:
+            import jwt
+        except ModuleNotFoundError:  # pragma: no cover
+            log.error("missing module: install FlaskSimpleAuth[jwt]")
+            raise
 
         try:
             data = jwt.decode(token, self._secret, leeway=self._grace * 60.0,
@@ -1296,11 +1365,13 @@ class FlaskSimpleAuth:
             scopes = data["scope"].split(" ") if "scope" in data else None
             return data["sub"], exp, scopes
         except jwt.ExpiredSignatureError:
-            log.debug(f"AUTH (jwt token): token {token} has expired")
-            raise self._Err("expired jwt auth token", 401)
+            if self._mode >= Mode.DEBUG:
+                log.debug(f"AUTH (jwt token): token has expired ({token})")
+            raise self._Err("expired jwt auth token: {token}", 401)
         except Exception as e:
-            log.debug(f"AUTH (jwt token): invalid token ({e})")
-            raise self._Err("invalid jwt token", 401, e)
+            if self._mode >= Mode.DEBUG:
+                log.debug(f"AUTH (jwt token): invalid token {token}: {e}")
+            raise self._Err("invalid jwt token: {token}", 401, e)
 
     def _get_any_token_auth_exp(self, token):
         """Return validated user and expiration, cached."""
@@ -1315,8 +1386,9 @@ class FlaskSimpleAuth:
         # must recheck token expiration
         now = dt.datetime.now(dt.timezone.utc) - dt.timedelta(minutes=self._grace)
         if now > exp:
-            log.debug(f"AUTH (token): token {token} has expired")
-            raise self._Err("expired auth token", 401)
+            if self._mode >= Mode.DEBUG:
+                log.debug(f"AUTH (token): token has expired ({token})")
+            raise self._Err("expired auth token: {token}", 401)
         self._local.scopes = scopes
         return user
 
@@ -1383,13 +1455,14 @@ class FlaskSimpleAuth:
                 lae = e
             except Exception as e:  # pragma: no cover
                 log.error(f"internal error in {a} authentication: {e}")
-                log.info(e, exc_info=True)
+                self._Exc(e)
 
         # even if not set, we say that the answer is the right one.
         self._local.source = "none"
 
         # rethrow last auth exception on failure
         if required and not self._local.user:
+            # we do not leak allowed authentication schemes
             raise lae or self._Err("missing authentication", 401)
 
         return self._local.user
@@ -1433,6 +1506,7 @@ class FlaskSimpleAuth:
     #  _group_authz: check group authorization
     #     _no_authz: validate that no authorization was needed
     #   _parameters: handle HTTP/JSON to python parameter translation
+    #     _noparams: just check that no parameters are passed
     #   _perm_authz: check per-object permissions
     #
     def _safe_call(self, path, level, fun, *args, **kwargs):
@@ -1441,9 +1515,10 @@ class FlaskSimpleAuth:
             return fun(*args, **kwargs)
         except ErrorResponse as e:  # something went wrong
             return self._Res(e.message, e.status)
-        except Exception as e:  # something went really wrong
+        except Exception as e:  # something went *really* wrong
             log.error(f"internal error on {request.method} {request.path}: {e}")
-            log.info(e, exc_info=True)
+            if self._Exc(e):
+                raise
             return self._Res(f"internal error caught at {level} on {path}", self._server_error)
 
     def _authenticate(self, path, auth=None):
@@ -1535,7 +1610,8 @@ class FlaskSimpleAuth:
                         return self._Res(e.message, e.status)
                     except Exception as e:
                         log.error(f"user_in_group failed: {e}")
-                        log.info(e, exc_info=True)
+                        if self._Exc(e):  # pragma: no cover
+                            raise
                         return self._Res("internal error in user_in_group", self._server_error)
                     if not isinstance(ok, bool):
                         log.error(f"type error in user_in_group: {ok}: {type(ok)}, must return a boolean")
@@ -1562,6 +1638,28 @@ class FlaskSimpleAuth:
 
         return decorate
 
+    # just check that there are no unused parameters
+    def _noparams(self, path):
+        def decorate(fun: Callable):
+            @functools.wraps(fun)
+            def wrapper(*args, **kwargs):
+                if self._reject_param:
+                    params = self._params()
+                    if params:
+                        sparams = set(params.keys())
+                        if not sparams.issubset(self._auth_params):
+                            bads = ' '.join(sorted(list(sparams - self._auth_params)))
+                            return f"unexpected {self._local.params} parameters: {bads}", 400
+                    fparams = request.files
+                    if fparams:
+                        return f"unexpected file parameters: {' '.join(sorted(fparams.keys()))}", 400
+
+                return self._safe_call(path, "no params", fun, *args, **kwargs)
+
+            return wrapper
+
+        return decorate
+
     def _parameters(self, path):
         """Decorator to handle request parameters."""
 
@@ -1608,23 +1706,25 @@ class FlaskSimpleAuth:
                 # NOTE *args and **kwargs are empty before being filled in from HTTP
 
                 # translate request parameters to named function parameters
-                params = self._params()
+                params = self._params()  # HTTP or JSON params
+                fparams = request.files  # FILE params
 
                 for p, typing in typings.items():
                     pn = names[p]
                     if typing in self._special_parameters:  # force specials
                         if p in params:
-                            return self._Res(f'unexpected parameter "{pn}"', 400)
-                        kwargs[p] = self._special_parameters[typing]()
+                            return self._Res(f'unexpected {self._local.params} parameter "{pn}"', 400)
+                        kwargs[p] = self._special_parameters[typing](pn)
                     elif p in kwargs:  # path parameter, or already seen?
                         if p in params:
-                            return self._Res(f'unexpected parameter "{pn}"', 400)
+                            return self._Res(f'unexpected {self._local.params} parameter "{pn}"', 400)
                         val = kwargs[p]
                         if not isinstance(val, types[p]):
                             try:
                                 kwargs[p] = typing(val)
                             except Exception as e:
-                                log.info(e, exc_info=True)
+                                if self._Exc(e):  # pragma: no cover
+                                    raise
                                 return self._Res(f'type error on path parameter "{p}": "{val}" ({e})', 400)
                     else:  # parameter not yet encountered
                         if pn in params:
@@ -1635,8 +1735,9 @@ class FlaskSimpleAuth:
                                 try:
                                     kwargs[p] = typing(val)
                                 except Exception as e:
-                                    log.info(e, exc_info=True)
-                                    return self._Res(f'type error on parameter "{pn}": "{val}" ({e})', 400)
+                                    if self._Exc(e):  # pragma: no cover
+                                        raise
+                                    return self._Res(f'type error on {self._local.params} parameter "{pn}": "{val}" ({e})', 400)
                             else:
                                 kwargs[p] = val
                         else:
@@ -1648,20 +1749,32 @@ class FlaskSimpleAuth:
                                 return self._Res(f'missing parameter "{pn}"', 400)
 
                 # possibly add others, without shadowing already provided ones
-                if keywords:
+                if keywords:  # handle **kwargs
                     for p in params:
                         if p not in kwargs:
-                            kwargs[p] = params[p]
+                            kwargs[p] = params[p]  # FIXME names[p]?
+                    for p in fparams:
+                        if p not in kwargs:
+                            kwargs[p] = fparams[p]  # FIXME?
                 elif self._mode >= Mode.DEBUG or self._reject_param:
                     # detect unused parameters and warn or reject them
                     for p in params:
-                        if p not in names and p not in self._names:
+                        # NOTE we silently ignore special authentication params
+                        if p not in names and p not in self._auth_params:
                             if self._mode >= Mode.DEBUG:
-                                log.debug(f"unexpected parameter {p} on {path}")
+                                log.debug(f"unexpected {self._local.params} parameter \"{p}\" on \"{path}\"")
                                 if self._mode >= Mode.DEBUG2:
                                     log.debug(debugParam())
                             if self._reject_param:
-                                return self._Res(f"unexpected parameter {p} on {path}", 400)
+                                return self._Res(f"unexpected {self._local.params} parameter \"{p}\" on \"{path}\"", 400)
+                    for p in fparams:
+                        if p not in names:
+                            if self._mode >= Mode.DEBUG:
+                                log.debug(f"unexpected file parameter \"{p}\" on \"{path}\"")
+                                if self._mode >= Mode.DEBUG2:
+                                    log.debug(debugParam())
+                            if self._reject_param:
+                                return self._Res(f"unexpected file parameter \"{p}\" on \"{path}", 400)
 
                 return self._safe_call(path, "parameters", fun, *args, **kwargs)
 
@@ -1713,7 +1826,8 @@ class FlaskSimpleAuth:
                         return self._Res(e.message, e.status)
                     except Exception as e:
                         log.error(f"internal error on {request.method} {request.path} permission {perm} check: {e}")
-                        log.info(e, exc_info=True)
+                        if self._Exc(e):  # pragma: no cover
+                            raise
                         return self._Res("internal error in permission check", self._server_error)
 
                     if ok is None:
@@ -1855,6 +1969,8 @@ class FlaskSimpleAuth:
             fun = self._perm_authz(newpath, first, *perms)(fun)
         if need_parameters:
             fun = self._parameters(newpath)(fun)
+        else:
+            fun = self._noparams(newpath)(fun)
         if groups:
             assert need_authenticate
             if auth == "oauth":
@@ -1877,7 +1993,7 @@ class FlaskSimpleAuth:
 
         assert fun != view_func, "some wrapping added"
 
-        # last wrapper to signal a routed function
+        # last wrapper to signal a "routed" function
         @functools.wraps(fun)
         def entry(*args, **kwargs):
             self._local.routed = True
