@@ -3,6 +3,7 @@
 # FIXME tests are not perfectly isolated as they should be…
 #
 
+import io
 import re
 import pytest
 import App
@@ -24,6 +25,8 @@ log = logging.getLogger("tests")
 # app._fsa._initialize()
 
 def check(code, res):
+    if res.status_code != code:
+        log.debug(f"BAD res = {res.data}")
     assert res.status_code == code
     return res
 
@@ -87,6 +90,7 @@ def push_auth(app, auth, token = None, carrier = None, name = None):
     app_saved_auth.update(a = app._auth, t = app._token, c = app._carrier, n = app._name)
     app._auth = [auth] if isinstance(auth, str) else auth
     app._token, app._carrier, app._name = token, carrier, name
+    app._auth_params.add(name)
 
 def pop_auth(app):
     d = app_saved_auth
@@ -261,7 +265,7 @@ def test_fsa_token():
         user = app._fsa._get_any_token_auth("Test:calvin:20201500000000:signature")
         assert False, "expecting a bad timestamp format"
     except fsa.ErrorResponse as e:
-        assert "unexpected limit" in e.message
+        assert "unexpected fsa token limit" in e.message
     # force expiration
     grace = app._fsa._grace
     app._fsa._grace = -1000000
@@ -431,7 +435,7 @@ def test_password_check(client):
     res = check(401, client.get("/read", data={"USER": "dad", "PASS": "bad-dad-password"}))
     assert b"invalid password for" in res.data
     res = check(401, client.get("/read", data={"USER": "dad"}))
-    assert b"missing password parameter" in res.data
+    assert b"missing param password parameter" in res.data
     pop_auth(fsa)
     # basic, through requests
     push_auth(fsa, ["basic"])
@@ -889,6 +893,7 @@ def test_per_route(client):
     PARAM = {"USER": "calvin", "PASS": App.UP["calvin"]}
     FAKE = {"LOGIN": "calvin"}
     token = app.create_token("calvin")
+    log.debug(f"calvin token: {token}")
     TOKEN = {"Authorization": f"Bearer {token}"}
     # basic
     log.debug("trying: basic")
@@ -1090,7 +1095,7 @@ def bad4():
 def test_bad_4(bad4):
     check(200, bad4.get("/ok"))
     res = check(500, bad4.get("/any"))
-    assert b"internal error caught at no authorization on /any" == res.data
+    assert b"internal error caught at no params on /any" == res.data
     check(404, bad4.get("/no-such-route"))
 
 
@@ -1125,9 +1130,9 @@ def test_bads():
     check(200, client.get("/youpi", data={"i": 5432}))
     check(200, client.get("/youpi", json={"i": 5432}))
     res = check(400, client.get("/youpi", data={"i": 5432, "j": "oops!"}))
-    assert b"unexpected parameter j" in res.data
+    assert b"unexpected http parameter \"j\"" in res.data
     res = check(400, client.get("/youpi", json={"i": 5432, "h": "oops!"}))
-    assert b"unexpected parameter h" in res.data
+    assert b"unexpected json parameter \"h\"" in res.data
 
 # per-object perms
 def test_object_perms(client):
@@ -1649,7 +1654,7 @@ def test_error_response():
     app.config.update(FSA_ERROR_RESPONSE="json", FSA_SECURE=False)
     app._fsa.initialize()
     # check that we take control of flask errors
-    app = fsa.Flask("not-implemented", FSA_ERROR_RESPONSE="json:bad")
+    app = fsa.Flask("not-implemented", FSA_LOGGING_LEVEL=logging.INFO, FSA_ERROR_RESPONSE="json:bad")
     @app.get("/implemented", authorize="ANY")
     def get_implemented():
         raise ErrorResponse("oops", 418)
@@ -1729,3 +1734,126 @@ def test_shadowing(client):
     assert res.data == b"Test: bla Manitoba"
     res = check(200, client.get("/shadow/bla?blup=Manitoba&blup=Quebec"))
     assert res.data == b"Test: bla Manitoba"
+
+def test_user_errors():
+    # check that user errors are raised again under FSA_KEEP_USER_ERRORS
+    class Oops(Exception):
+        pass
+    app = fsa.Flask("user-errors", FSA_KEEP_USER_ERRORS=True)
+    @app.get("/oops", authorize="ANY")
+    def get_oops():
+        raise Oops("internal error in get_oops!")
+    client = app.test_client()
+    res = check(500, client.get("/oops"))
+    # lengthy flask-generated message
+    assert b"The server encountered an internal error" in res.data
+    # without rethrow
+    app._fsa._keep_user_errors = False
+    res = check(500, client.get("/oops"))
+    # to-the-point fsa-generated message
+    assert b"internal error caught at" in res.data
+
+def test_param_params():
+    app = fsa.Flask("pp",
+        FSA_MODE="debug2",
+        FSA_DEBUG_LEVEL=logging.DEBUG,
+        FSA_AUTH="param",
+        FSA_PARAM_USER="login",
+        FSA_PARAM_PASS="password"
+    )
+
+    CALVIN = {"login": "calvin", "password": "hobbes"}
+
+    USERS = {
+        "calvin": app.hash_password("hobbes"),
+        "hobbes": app.hash_password("calvin"),
+    }
+
+    @app.get_user_pass
+    def get_user_pass(login: str):
+        return USERS[login] if login in USERS else None
+
+    # note: "login" and "password" parameters must be ignored
+    @app.post("/log0", authorize="ALL")
+    def post_log0():
+        return f"current user is {app.get_user()}", 200
+
+    # note: "login" and "password" parameters must be ignored
+    @app.post("/log1", authorize="ALL")
+    def post_log1(hello: str):
+        return f"current user is {app.get_user()}", 200
+
+    @app.post("/log2", authorize="ALL")
+    def post_log2(login: fsa.CurrentUser, password: str = "world!"):
+        return f"login={login} hello={password}", 200
+
+    @app.post("/log3", authorize="ALL")
+    def post_log5(login: str, password: str):
+        return f"login={login} password={password}", 200
+
+    client = app.test_client()
+
+    # missing auth parameters
+    res = check(401, client.post("/log0", data={}))
+    res = check(401, client.post("/log0", data={"login": "calvin"}))
+    res = check(401, client.post("/log0", json={"login": "calvin"}))
+    res = check(401, client.post("/log0", data={"password": "hobbes"}))
+    res = check(401, client.post("/log0", json={"password": "hobbes"}))
+    # unexpected "foo" parameter
+    res = check(400, client.post("/log0", data={"login": "calvin", "password": "hobbes", "foo": "bla"}))
+    res = check(400, client.post("/log0", json={"login": "calvin", "password": "hobbes", "foo": "bla"}))
+
+    # OK
+    res = check(200, client.post("/log0", data=CALVIN))
+    assert b"current user is calvin" == res.data
+
+def test_file_storage():
+
+    def bfile(contents: bytes, name: str = "foo.txt", ct: str = "text/plain"):
+        return (io.BytesIO(contents), name, ct)
+
+    app = fsa.Flask("file-storage", FSA_MODE="debug2")
+
+    @app.post("/upload", authorize="ANY")
+    def post_upload(file: fsa.FileStorage):
+        return f"file={file.filename}", 201
+
+    @app.post("/uploads", authorize="ANY")
+    def post_uploads(**kwargs):
+        return " ".join(sorted(kwargs.keys())), 201
+
+    @app.post("/mix", authorize="ANY")
+    def post_mix(data: int, file: fsa.FileStorage):
+        return f"data={data} file={file.filename}", 201
+
+    @app.post("/empty", authorize="ANY")
+    def post_empty():
+        return "nothing", 200
+
+    client = app.test_client()
+
+    # /upload
+    res = check(201, client.post("/upload", data={"file": bfile(b"hello file!\n")}))
+    assert b"file=foo.txt" in res.data
+    res = check(400, client.post("/upload", data={"stuff": bfile(b"hello stuff!\n")}))
+    assert b"missing file parameter \"file\"" in res.data
+    res = check(400, client.post("/upload", data={"file": bfile(b"hello file!\n"), "stuff": bfile(b"hello stuff!\n")}))
+    assert b"unexpected file parameter \"stuff\"" in res.data
+    res = check(400, client.post("/upload", data={"file": "bla.txt"}))
+    assert b"unexpected http parameter \"file\"" in res.data
+
+    # /uploads
+    res = check(201, client.post("/uploads", data={"foo": bfile(b"hello foo!\n"), "bla": bfile(b"hello bla!\n")}))
+    assert b"bla foo" in res.data
+
+    # /mix
+    res = check(201, client.post("/mix", data={"data": 42, "file": bfile(b"hello file!\n")}))
+    assert b"data=42 file=foo.txt" in res.data
+    res = check(400, client.post("/mix", data={"data": bfile(b"hello data!\n"), "file": bfile(b"hello file!\n")}))
+    assert b"missing parameter \"data\"" in res.data
+
+    # /empty
+    res = check(200, client.post("/empty"))
+    assert b"nothing" in res.data
+    res = check(400, client.post("/empty", data={"foo": bfile(b"hello foo!\n"), "bla": bfile(b"hello bla!\n")}))
+    assert b"unexpected file parameters: bla foo" in res.data
